import collections

from enum import *



class TreeNode:
    def __init__(self, language_element, subnodes):
        self.language_element = language_element
        self.subnodes = subnodes

    def __str__(self):
        return str(self.language_element) + " " + str(self.subnodes)


def walk_the_tree(tree, level = 1):
    if tree == None:
        return

    print("    " * (level -1) + "|---" + str(tree.language_element) + " with " + str(len(tree.subnodes)) + " subnodes")
    for subnode in tree.subnodes:
        walk_the_tree(subnode, level + 1)




CFG = collections.namedtuple("CFG", "nonterminals terminals rules start")
# We define a context-free grammar as:
# 1. A list of nonterminal symbols
# 2. A list of terminal symbols
# 3. A list of production rules
# 4. A start symbol.

# Each symbol is defined by a positive integer.


# For the ETF expression grammar:

# Terminal symbols are:
# INTEGER, ADDOP, MULTOP, OPENPAREN, CLOSEPAREN

# Nonterminal symbols are:
# EXPRESSION, TERM, FACTOR

# The start symbol is EXPRESSION

# The rules are:

# EXPRESSION -> EXPRESSION ADDOP TERM
# EXPRESSION -> TERM
# TERM -> TERM MULTOP FACTOR
# TERM -> FACTOR
# FACTOR -> OPENPAREN EXPRESSION CLOSEPAREN
# FACTOR -> INTEGER

symbols = Enum("Symbols", "EXPRESSION TERM FACTOR INTEGER ADDOP MULTOP OPENPAREN CLOSEPAREN")
start = symbols.EXPRESSION



# The rules are a list of tuples, each of which represents a rule, as follows:

# (Nonterminal symbol, [ordered list of symbols (terminal or nonterminal) that are reduced to aforementioned nonterminal])

rules = [
         (symbols.EXPRESSION, [symbols.EXPRESSION, symbols.ADDOP,      symbols.TERM      ]),
         (symbols.EXPRESSION, [symbols.TERM                                              ]),
         (symbols.TERM,       [symbols.TERM,       symbols.MULTOP,     symbols.FACTOR    ]),
         (symbols.TERM,       [symbols.FACTOR                                            ]),
         (symbols.FACTOR,     [symbols.OPENPAREN,  symbols.EXPRESSION, symbols.CLOSEPAREN]),
         (symbols.FACTOR,     [symbols.INTEGER                                           ])
         ]

print(rules)



# Once we have a CFG's definition, we can use Boltzmann sampling in order to generate random strings
# from that CFG.


# This, however, is not Boltzmann sampling.

# We examine all the leaves on the current derivation tree. Those mark where we possibly could apply a rule.
# We then determine which leaves are amenable to rule application, and then we randomly select a leaf and for
# that leaf, we randomly select a rule, and apply it.

# To avoid being accidentally quadratic, we cache a list of the leaves, and also cache the current length of
# the generated string (in order to do pseudo-Boltzmann sampling).



def apply_rule(tree, leafcache, cache_index, ruleset, rule_number):
    # first, we isolate the place on the tree that we are going to do the graft.
    opsite = tree

    location = leafcache[cache_index]

    for depth_index in location:
        opsite = opsite.subnodes[depth_index]
        print("TMP index ", depth_index, " gives ",opsite)

    print(opsite)
    walk_the_tree(opsite)

    # now we verify that we can legitimately apply the given rule here.

    the_rule = ruleset[rule_number]

    print("We will try applying the following rule:", the_rule)

    # We verify that the item at the operation site is:
    # 1) a leaf (it has no subnodes)
    # and
    # 2) has the correct value

    print("Opsite language value", opsite.language_element)
    print("Rule expects", the_rule[0])

    assert(opsite.language_element == the_rule[0])

    assert(opsite.subnodes == [])

    print("opsite has no subnodes, all good")

    new_leaves = []

    for idx, new_subnode in enumerate(the_rule[1]):
        print(new_subnode)
        x = TreeNode(new_subnode, [])
        opsite.subnodes.append(x)

        leafcache.append(location + [idx])

    print("Now the opsite looks like")
    walk_the_tree(opsite)

    # update the leaf cache

    del leafcache[cache_index]






test_tree = TreeNode(symbols.EXPRESSION, [
                        TreeNode(symbols.ADDOP,[]),
                        TreeNode(symbols.MULTOP,[
                                    TreeNode(symbols.EXPRESSION,[]), TreeNode(symbols.TERM,[]),TreeNode(symbols.FACTOR,[])
                                   ])
                        ])

walk_the_tree(test_tree)


#def next_stage(derivation_tree_with_cached, language_definition):

leafcache = [[1,1]]
apply_rule(test_tree,leafcache,0,rules,2)

print("and the tree looks like")
walk_the_tree(test_tree)

print("updated leafcache")
print(leafcache)

# Furthermore, we also note that the description of a context-free grammar is *itself* context-free
# so if we take the CFG-description grammar (BNF or something isomorphic to it), and use Boltzmann
# sampling on *it*, we will generate candidate grammars; which will be used to test the FPGA-based
# parser generator against a reference LR parser generator implementation. Naturally, this procedure 
# may not produce LR-parseable grammars or even deterministic context-free grammars; but if the grammar
# is rejected by both the FPGA parser generator and reference parser generator; this is fine, we just
# move on to the next one.


# Once we randomly generate a CFG, we randomly generate strings that are valid strings, but also
# generate exemplars that are potentially *invalid* strings (both by mutating valid strings but
# also by generating random sequences ab initio from the terminal symbols). The potentially invalid
# strings are tested against the reference parser, those that are accidentally valid are discarded,
# and those that are invalid are used in the automatic equivalence testing.


# We fail our candidate parser generator if:
#
# 1. During the generation phase, it rejects a grammar the reference generator doesn't reject, or
#    accepts a grammar the reference generator rejects.

# 2. During the execution phase, it rejects a string the reference generated parser doesn't reject,
#    or accepts a string the reference generated parser rejects.

# 3. During the execution phase, it accepts a string that the reference generated parser accepts; but
#    the parse tree it generates is not identical to the parse tree that the reference generated parser
#    produced.



